#!/usr/bin/env wish

#******************************
# 
# osBackup
# Version 2022-08-26
# Copyright Oliver Scholl < tcltk@posteo.de > < www.tcltk.info >
#
# License:
# This program is free software; you can redistribute it and/or 
# modify it under the terms of the GNU General Public License as 
# published by the Free Software Foundation; either version 2 of 
# the license, or any later version.
#
# This program is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# merchantability or fitness for a particular purpose. 
# See the GNU General Public License for more details.
#
#*************************************

#*** Generelle Einstellungen ***

set tclexecomp 0 ; # 1=wenn das Programm mit tclexecomp benutzt wird. Sonst 0.

package require msgcat
package require sqlite3
package require md5

# Workaround fuer tclexecomp
namespace inscope :: [list ::source -encoding utf-8  [file join [file dirname [info script]] de.msg]]
namespace inscope :: [list ::source -encoding utf-8  [file join [file dirname [info script]] en.msg]]
# source /zvfs/de.msg

font create GlobSchrift -family Helvetica -size 12 -slant roman -weight normal
ttk::style theme use default
ttk::style configure TButton -font GlobSchrift
ttk::style configure TLabel -font GlobSchrift
ttk::style configure TLabelFrame -font GlobSchrift 
ttk::style configure TMenubutton -font GlobSchrift 
ttk::style configure TRadiobutton -font GlobSchrift
ttk::style configure TEntry -font GlobSchrift 
ttk::style configure TCombobox -font GlobSchrift 
ttk::style configure TCheckbutton -font GlobSchrift
ttk::style configure TNotebook.Tab -font GlobSchrift 
ttk::style configure Treeview -font GlobSchrift 
ttk::style configure Treeview.Heading -font GlobSchrift
option add *TCombobox*Listbox.font GlobSchrift
option add *Dialog.msg.font GlobSchrift

set GlobKonf(Skriptname) [info script]
if {[file type $GlobKonf(Skriptname)] == "link"} {
	set GlobKonf(Skriptname) [file readlink $GlobKonf(Skriptname)]
}
::msgcat::mclocale de ;# Alternative Sprachen: de=deutsch en=englisch
::msgcat::mcload [file join [file dirname $GlobKonf(Skriptname)]]

if {$tclexecomp == 1} {
	set GlobKonf(Konfigurationsdatei) [file join [file dirname [file normalize [info nameofexecutable]]] Konf.txt] 
} else {
	set GlobKonf(Konfigurationsdatei) [file join [file dirname [file normalize [info script]]] Konf.txt] 
}

#*** Einstellungen zum Programm ***

set GlobKonf(ProgrammName) "osBackup"
set GlobKonf(ProgrammVersion) "2022-08-26"
set GlobKonf(ProgrammTitel) "Backup"

set Status "" ; # Statusmeldung

array set Protokoll {} ; # Backup-Protokoll
set Protokoll(QuelleDubletten) ""
set Protokoll(ZielDubletten) ""
set Protokoll(ZielNeueOrdner) ""
set Protokoll(ZielDateiLoeschen) ""
set Protokoll(DateiKopieren) ""
set Protokoll(DateiVerschieben) ""
set Protokoll(ZielOrdnerLoeschen) ""

set Zeit "" ; # maximale Kopierzeit. Danach stoppt das Programm

set GlobKonf(Quelle) $env(HOME)
set GlobKonf(Ziel) $env(HOME)
set GlobKonf(Sprache) "deutsch"
set GlobKonf(Schriftgroesse) 12
set GlobKonf(Md5sum) 0

if {[file exist $GlobKonf(Konfigurationsdatei)]} {
	set f [open $GlobKonf(Konfigurationsdatei) r]
	while {[gets $f Zeile] >= 0} {
		set Nummer [string rang $Zeile 0 0]
		set Wert [string rang $Zeile 2 end]
		switch $Nummer {
			1 {set GlobKonf(Quelle) $Wert}
			2 {set GlobKonf(Ziel) $Wert}
			3 {set GlobKonf(Sprache) $Wert}
			4 {set GlobKonf(Schriftgroesse) $Wert}
			5 {set GlobKonf(Md5sum) $Wert}
		}
	}
	close $f
}

#*** Prozeduren ***

proc DatenbankAnlegen {} {
	# Erzeugt einen neuen Befehl namens Datenbank

	sqlite3 Datenbank ":memory:"

	Datenbank eval {create table tQuelleDateien( \
		ID integer primary key autoincrement, \
		Datei text, \
		Pfad text, \
		PfadRelativ text, \
		DateinameOhnePfad text, \
		Fingerabdruck text
	)}
	
	Datenbank eval {create table tQuelleOrdner( \
		ID integer primary key autoincrement, \
		Ordner text, \
		OrdnerRelativ text
	)}

	Datenbank eval {create table tZielDateien( \
		ID integer primary key autoincrement, \
		Datei text, \
		Pfad text, \
		PfadRelativ text, \
		DateinameOhnePfad text, \
		Fingerabdruck text
	)}

	Datenbank eval {create table tZielOrdner( \
		ID integer primary key autoincrement, \
		Ordner text, \
		OrdnerRelativ text
	)}
	
	#Index erzeugen
	Datenbank eval {create index QuelleDateiIndex on tQuelleDateien(Datei)}
	Datenbank eval {create index QuelleFingerabdruckIndex on tQuelleDateien(Fingerabdruck)}
	Datenbank eval {create index QuellePfadRelativIndex on tQuelleDateien(PfadRelativ)}
	
	#Index erzeugen
	Datenbank eval {create index ZielDateiIndex on tZielDateien(Datei)}
	Datenbank eval {create index ZielFingerabdruckIndex on tZielDateien(Fingerabdruck)}
	Datenbank eval {create index ZielPfadRelativIndex on tZielDateien(PfadRelativ)}

	#Index erzeugen
	Datenbank eval {create index QuelleOrdnerIndex on tQuelleOrdner(Ordner)}
	Datenbank eval {create index QuelleOrdnerRelativIndex on tQuelleOrdner(OrdnerRelativ)}

	#Index erzeugen
	Datenbank eval {create index ZielOrdnerIndex on tZielOrdner(Ordner)}
	Datenbank eval {create index ZielOrdnerRelativIndex on tZielOrdner(OrdnerRelativ)}
}

proc DatenbankLeeren {} {
	# die Datenbank leeren

	Datenbank eval {delete from tQuelleDateien}
	Datenbank eval {delete from tQuelleOrdner}
	Datenbank eval {delete from tZielDateien}
	Datenbank eval {delete from tZielOrdner}
}

proc DateienEinlesen {Startordner} {
	global ListeDateien
	global ListeOrdner
	global Status
	global Protokoll
	
	# Liest alle Ordner und Dateien ein 
	set IndexOrdner 0
	set ListeOrdner {}
	set ListeDateien {}

	lappend ListeOrdner $Startordner

	set FortschrittZaehler 0
	set FortschrittZaehler2 0
	set Status [::msgcat::mc "Bitte warten: %s" $FortschrittZaehler2]
	update idletasks
	
	while {$IndexOrdner < [llength $ListeOrdner] } {
		set Ordner [lindex $ListeOrdner $IndexOrdner]

		incr FortschrittZaehler
		if {$FortschrittZaehler>=100} {
			set FortschrittZaehler 0
			incr FortschrittZaehler2 100
			set Status [::msgcat::mc "Bitte warten: %s" $FortschrittZaehler2]
			update idletasks
		}
		if {[string index $Ordner end] != "."} {
			# Nicht die Ordner . bzw .. aufrufen
			if { [catch {glob -nocomplain -directory $Ordner -types {f} *} ret] == 0 } {
				foreach i $ret {
					lappend ListeDateien $i
				}
			}
			if { [catch {glob -nocomplain -directory $Ordner -types {f hidden} *} ret] == 0 } {
				foreach i $ret {
					lappend ListeDateien $i
				}
			}
			if { [catch {glob -nocomplain -directory $Ordner -types {d} *} ret] == 0 } {
				foreach i $ret {
					if {[file tail $i] ne "." && [file tail $i] ne ".."} {
						lappend ListeOrdner $i
					}
				}
			}
			if { [catch {glob -nocomplain -directory $Ordner -types {d hidden} *} ret] == 0 } {
				foreach i $ret {
					if {[file tail $i] ne "." && [file tail $i] ne ".."} {
						lappend ListeOrdner i
					}
				}
			}
		}
		incr IndexOrdner +1
	}
	set Status [::msgcat::mc "Bereit."]
	update idletasks
}

proc SpracheAendern {Sprache} {
	# stellt die Sprache ein
	global GlobKonf

	switch $Sprache {
		Deutsch {
			set GlobKonf(Sprache) "Deutsch"
			::msgcat::mclocale de
		}
		English {
			set GlobKonf(Sprache) "English"
			::msgcat::mclocale en
		}
	}
	::msgcat::mcload [file join [file dirname $GlobKonf(Skriptname)]]

	.m entryconfigure 0 -label [::msgcat::mc "Schliessen"]
	.m entryconfigure 1 -label [::msgcat::mc "Sprache"]
	.m entryconfigure 2 -label [::msgcat::mc "Schriftgroesse"]
	.m entryconfigure 3 -label [::msgcat::mc "Info"]

	.f1.lbQuelle configure -text [::msgcat::mc "Quelle:"]
	.f1.lbZiel configure -text [::msgcat::mc "Ziel:"]
	.f1.cbMd5sum configure -text [::msgcat::mc "Md5sum verwenden"]
	.f1.btAnalyse configure -text [::msgcat::mc "Analyse"]
	.f1.btStart configure -text [::msgcat::mc "Start"]
	.f1.lbZeit configure -text [::msgcat::mc "maximale Zeit (in Minuten):"]

	.n tab .n.f1 -text [::msgcat::mc "Hinweise"]
	.n tab .n.f2 -text [::msgcat::mc "Ordner"]
	.n tab .n.f3 -text [::msgcat::mc "Dateien"]

	.n.f1.lbProtokollQuelleDubletten configure -text [::msgcat::mc "Folgende Dateien kommen in der Quelle mehrfach vor:"]
	.n.f1.lbProtokollZielDubletten configure -text [::msgcat::mc "Folgende Dateien kommen im Ziel mehrfach vor:"]
	.n.f2.lbProtokollZielNeueOrdner configure -text [::msgcat::mc "Folgende Ordner werden im Ziel neu angelegt:"]
	.n.f3.lbProtokollZielDateiLoeschen configure -text [::msgcat::mc "Folgende Dateien werden im Ziel geloescht:"]
	.n.f3.lbProtokollDateiKopieren configure -text [::msgcat::mc "Folgende Dateien werden kopiert:"]
	.n.f3.lbProtokollDateiVerschieben configure -text [::msgcat::mc "Folgende Dateien werden im Ziel verschoben:"]
	.n.f2.lbProtokollZielOrdnerLoeschen configure -text [::msgcat::mc "Folgende Ordner werden im Ziel geloescht:"]
		
	update idletasks
}

proc Md5sumErzeugenTCL {Datei} {
	# Benutzt den TCL-eigenen Befehl.
	# erzeugt eine md5sum der Datei
	# Rückgabe: md5sum-Wert (hexadezimal)
	set Md5sum ""
	catch {
		set Md5sum [md5::md5 -hex -file $Datei]]
	}
	return $Md5sum
}

proc Md5sumErzeugen {Datei} {
	# Benutzt den Linux-Befehl.
	# erzeugt eine md5sum der Datei
	# Rückgabe: md5sum-Wert (hexadezimal)
	set Md5sum ""
	catch {
		set Md5sum [exec md5sum $Datei]
	}
	set tmp [string first " " $Md5sum]
	set Md5sum [string range $Md5sum 0 [expr $tmp -1]]
	return $Md5sum
}

proc FingerabdruckErstellen {Datei} {
	global GlobKonf
	
	set NameMitEndung [file tail $Datei]
	if {$GlobKonf(Md5sum) == 1} {
		set Fingerabdruck [Md5sumErzeugenTCL $Datei]
	} else {
		set tmpZeit ""
		set GroesseBytes ""
		catch {set tmpZeit [file mtime $Datei]}
		set tmpZeit [expr int($tmpZeit/10)] ; # manchmal haben identische Dateien eine um 1 Sekunde verschobene Zeitangabe. Deshalb die Sekunde weglassen.	
		catch {set GroesseBytes [file size $Datei]}
		set Fingerabdruck "$NameMitEndung-$tmpZeit-$GroesseBytes"
	}
	return $Fingerabdruck
}

proc FingerabdruckErstellenQuelleDublette {Datei} {
	global GlobKonf

	set Dateiname [string range $Datei [string length $GlobKonf(Quelle)] end]
			
	if {$GlobKonf(Md5sum) == 1} {
		set Fingerabdruck [Md5sumErzeugenTCL $Datei]
		
	} else {
		set tmpZeit ""
		set GroesseBytes ""
		catch {set tmpZeit [file mtime $Datei]}
		set tmpZeit [expr int($tmpZeit/10)] ; # manchmal haben identische Dateien eine um 1 Sekunde verschobene Zeitangabe. Deshalb die Sekunde weglassen.	
		catch {set GroesseBytes [file size $Datei]}
		set Fingerabdruck "$tmpZeit-$GroesseBytes"
	}
	set Fingerabdruck "$Dateiname-$Fingerabdruck"
	return $Fingerabdruck
}

proc FingerabdruckErstellenZielDublette {Datei} {
	global GlobKonf

	set Dateiname [string range $Datei [string length $GlobKonf(Ziel)] end]
			
	if {$GlobKonf(Md5sum) == 1} {
		set Fingerabdruck [Md5sumErzeugenTCL $Datei]
		
	} else {
		set tmpZeit ""
		set GroesseBytes ""
		catch {set tmpZeit [file mtime $Datei]}
		set tmpZeit [expr int($tmpZeit/10)] ; # manchmal haben identische Dateien eine um 1 Sekunde verschobene Zeitangabe. Deshalb die Sekunde weglassen.	
		catch {set GroesseBytes [file size $Datei]}
		set Fingerabdruck "$tmpZeit-$GroesseBytes"
	}
	set Fingerabdruck "$Dateiname-$Fingerabdruck"
	return $Fingerabdruck
}

# ***** Prozeduren, die andere Prozeduren aufrufen *****

proc DateienInDatenbankSpeichern_Quelle {} {
	# Dateien in Tabelle speichern
	global GlobKonf
	global ListeDateien
	global Status
	
	set FortschrittZaehler 0
	set FortschrittZaehler2 0
	set Status [::msgcat::mc "Bitte warten: %s" $FortschrittZaehler2]
	update idletasks

	set i 0
	set ListeDateien [lsort -unique $ListeDateien]
	Datenbank eval {BEGIN TRANSACTION}		
	foreach Eintrag $ListeDateien {
		incr FortschrittZaehler
		if {$FortschrittZaehler>=100} {
			set FortschrittZaehler 0
			incr FortschrittZaehler2 100
			set Status [::msgcat::mc "Bitte warten: %s" $FortschrittZaehler2]
			update idletasks
		}
		
		set Pfad [file dirname $Eintrag]
		set PfadRelativ [string range $Pfad [string length $GlobKonf(Quelle)] end]
	
		set NameMitEndung [file tail $Eintrag]

		if {$GlobKonf(Md5sum) == 1} {
			set Fingerabdruck [Md5sumErzeugenTCL $Eintrag]
		} else {
			set Fingerabdruck [FingerabdruckErstellen $Eintrag]
		}
		incr i
		Datenbank eval {insert into tQuelleDateien values(NULL,$Eintrag,$Pfad,$PfadRelativ,$NameMitEndung,$Fingerabdruck)}
	}
	Datenbank eval {END TRANSACTION}		
	#puts [Datenbank eval {select * from tQuelleDateien}]

	set Status [::msgcat::mc "Bereit."]
	update idletasks
}

proc DateienInDatenbankSpeichern_Ziel {} {
	# Dateien Tabelle speichern
	global GlobKonf
	global ListeDateien
	global Status
	
	set FortschrittZaehler 0
	set FortschrittZaehler2 0
	set Status [::msgcat::mc "Bitte warten: %s" $FortschrittZaehler2]
	update idletasks

	set i 0
	set ListeDateien [lsort -unique $ListeDateien]
	Datenbank eval {BEGIN TRANSACTION}		
	foreach Eintrag $ListeDateien {
		incr FortschrittZaehler
		if {$FortschrittZaehler>=100} {
			set FortschrittZaehler 0
			incr FortschrittZaehler2 100
			set Status [::msgcat::mc "Bitte warten: %s" $FortschrittZaehler2]
			update idletasks
		}

		set Pfad [file dirname $Eintrag]
		set PfadRelativ [string range $Pfad [string length $GlobKonf(Ziel)] end]

		set NameMitEndung [file tail $Eintrag]

		if {$GlobKonf(Md5sum) == 1} {
			set Fingerabdruck [Md5sumErzeugenTCL $Eintrag]
		} else {
			set Fingerabdruck [FingerabdruckErstellen $Eintrag]
		}
		incr i
		Datenbank eval {insert into tZielDateien values(NULL,$Eintrag,$Pfad,$PfadRelativ,$NameMitEndung,$Fingerabdruck)}
	}
	Datenbank eval {END TRANSACTION}		
	#puts [Datenbank eval {select * from tZielDateien}]

	set Status [::msgcat::mc "Bereit."]
	update idletasks
}

proc OrdnerInDatenbankSpeichern_Quelle {} {
	# Ordner in Tabelle speichern
	global GlobKonf
	global ListeOrdner
	global Status
	
	set FortschrittZaehler 0
	set FortschrittZaehler2 0
	set Status [::msgcat::mc "Bitte warten: %s" $FortschrittZaehler2]
	update idletasks

	set i 0
	set ListeOrdner [lsort -unique $ListeOrdner]
	Datenbank eval {BEGIN TRANSACTION}		
	foreach Eintrag $ListeOrdner {
		incr FortschrittZaehler
		if {$FortschrittZaehler>=100} {
			set FortschrittZaehler 0
			incr FortschrittZaehler2 100
			set Status [::msgcat::mc "Bitte warten: %s" $FortschrittZaehler2]
			update idletasks
		}
		set OrdnerRelativ [string range $Eintrag [string length $GlobKonf(Quelle)] end]

		Datenbank eval {insert into tQuelleOrdner values(NULL,$Eintrag,$OrdnerRelativ)}
	}
	Datenbank eval {END TRANSACTION}		
	#puts [Datenbank eval {select * from tQuelleOrdner}]

	set Status [::msgcat::mc "Bereit."]
	update idletasks
}

proc OrdnerInDatenbankSpeichern_Ziel {} {
	# Ordner in Tabelle speichern
	global GlobKonf
	global ListeOrdner
	global Status
	
	set FortschrittZaehler 0
	set FortschrittZaehler2 0
	set Status [::msgcat::mc "Bitte warten: %s" $FortschrittZaehler2]
	update idletasks

	set i 0
	set ListeOrdner [lsort -unique $ListeOrdner]
	Datenbank eval {BEGIN TRANSACTION}		
	foreach Eintrag $ListeOrdner {
		incr FortschrittZaehler
		if {$FortschrittZaehler>=100} {
			set FortschrittZaehler 0
			incr FortschrittZaehler2 100
			set Status [::msgcat::mc "Bitte warten: %s" $FortschrittZaehler2]
			update idletasks
		}
		set OrdnerRelativ [string range $Eintrag [string length $GlobKonf(Ziel)] end]

		Datenbank eval {insert into tZielOrdner values(NULL,$Eintrag,$OrdnerRelativ)}
	}
	Datenbank eval {END TRANSACTION}		
	#puts [Datenbank eval {select * from tZielOrdner}]

	set Status [::msgcat::mc "Bereit."]
	update idletasks
}

proc BackupStarten {{Start 0}} {	
	# Start = Boolean. 1=Backup starten, 0=nur analysieren

	global GlobKonf
	global Status
	global Protokoll
	global Zeit

	set Protokoll(QuelleDubletten) ""
	set Protokoll(ZielDubletten) ""
	set Protokoll(ZielNeueOrdner) ""
	set Protokoll(ZielDateiLoeschen) ""
	set Protokoll(DateiKopieren) ""
	set Protokoll(DateiVerschieben) ""
	set Protokoll(ZielOrdnerLoeschen) ""

	set Abbruch 0 ; # Boolean. 1 = vorzeitiges Ende wegen Zeitüberschreitung; 2 = ein Fehler ist aufgetreten
	
	set Cursor [. cget -cursor]
	. configure -cursor watch

	if {$Start == 0} {
		# Quelle einlesen
		DateienEinlesen $GlobKonf(Quelle)
		DateienInDatenbankSpeichern_Quelle
		OrdnerInDatenbankSpeichern_Quelle
		
		# Ziel einlesen
		DateienEinlesen $GlobKonf(Ziel)
		DateienInDatenbankSpeichern_Ziel
		OrdnerInDatenbankSpeichern_Ziel
	}
	
	set Status [::msgcat::mc "in Arbeit"]
	update idletasks
	
	# Prüfen, dass in der Quelle keine Dubletten (bzgl. des Fingerabdrucks) vorkommen
	set tmp [Datenbank eval {SELECT DISTINCT tQuelleDateien.Fingerabdruck FROM tQuelleDateien GROUP BY tQuelleDateien.Fingerabdruck HAVING (count(tQuelleDateien.Fingerabdruck) > 1)}]
	if {$tmp != ""} {
		set tmp [lsort $tmp]
		Datenbank eval {BEGIN TRANSACTION}		
		foreach Element $tmp {
			set tmp2 [Datenbank eval {SELECT DISTINCT tQuelleDateien.Datei FROM tQuelleDateien WHERE tQuelleDateien.Fingerabdruck=$Element}]
			set tmp2 [lsort $tmp2]

			# für die Gui-Ausgabe ein = zwischen die Dateiname einfügen
			set tmp3 ""
			if {[llength $tmp2] > 1} {
				foreach Element3 $tmp2 {
					append tmp3 $Element3
					append tmp3 " = "
				}
				set tmp3 [string range $tmp3 0 end-3]
			} else {
				set tmp3 $tmp2
			}

			lappend Protokoll(QuelleDubletten) $tmp3
			foreach Element2 $tmp2 {
				set Fingerabdruck2 [FingerabdruckErstellenQuelleDublette $Element2]
				set tmp2 [Datenbank eval {UPDATE tQuelleDateien SET Fingerabdruck=$Fingerabdruck2 WHERE tQuelleDateien.Datei=$Element2}]
			}
		}
		Datenbank eval {END TRANSACTION}		
	}

	# Prüfen, dass im Ziel keine Dubletten (bzgl. des Fingerabdrucks) vorkommen
	set tmp [Datenbank eval {SELECT DISTINCT tZielDateien.Fingerabdruck FROM tZielDateien GROUP BY tZielDateien.Fingerabdruck HAVING (count(tZielDateien.Fingerabdruck) > 1)}]
	if {$tmp != ""} {
		set tmp [lsort $tmp]
		Datenbank eval {BEGIN TRANSACTION}		
		foreach Element $tmp {
			set tmp2 [Datenbank eval {SELECT DISTINCT tZielDateien.Datei FROM tZielDateien WHERE tZielDateien.Fingerabdruck=$Element}]
			set tmp2 [lsort $tmp2]

			# für die Gui-Ausgabe ein = zwischen die Dateiname einfügen
			set tmp3 ""
			if {[llength $tmp2] > 1} {
				foreach Element3 $tmp2 {
					append tmp3 $Element3
					append tmp3 " = "
				}
				set tmp3 [string range $tmp3 0 end-3]
			} else {
				set tmp3 $tmp2
			}

			lappend Protokoll(ZielDubletten) $tmp3
			foreach Element2 $tmp2 {
				set Fingerabdruck2 [FingerabdruckErstellenZielDublette $Element2]
				set tmp2 [Datenbank eval {UPDATE tZielDateien SET Fingerabdruck=$Fingerabdruck2 WHERE tZielDateien.Datei=$Element2}]
			}
		}
		Datenbank eval {END TRANSACTION}		
	}

	# Alle Ordner aus der Quelle, die nicht im Ziel vorkommen
	# Diese Ordner müssen angelegt werden.
	set tmp [Datenbank eval {SELECT DISTINCT tQuelleOrdner.OrdnerRelativ FROM tQuelleOrdner LEFT OUTER JOIN tZielOrdner ON tQuelleOrdner.OrdnerRelativ=tZielOrdner.OrdnerRelativ WHERE tZielOrdner.ID IS NULL}]
	set tmp [lsort $tmp]
	foreach Element $tmp {
		lappend Protokoll(ZielNeueOrdner) $GlobKonf(Ziel)$Element
		if {$Start == 1} {
			catch {file mkdir $GlobKonf(Ziel)$Element}
			catch {exec sync}
			if {![file exist $GlobKonf(Ziel)$Element]} {
				tk_messageBox -title [::msgcat::mc "Fehler"] -message [::msgcat::mc "Der Ordner %s kann nicht erstellt werden." $GlobKonf(Ziel)$Element] -icon question -type ok -default ok
				set Abbruch 2
				break
			}
		}
	}
	
	if {$Abbruch == 0} {
		# Alle Dateien aus dem Ziel, die nicht in der Quelle vorkommen
		# Diese Dateien müssen gelöscht werden.
		set tmp [Datenbank eval {SELECT DISTINCT tZielDateien.Datei FROM tZielDateien LEFT OUTER JOIN tQuelleDateien ON tZielDateien.Fingerabdruck=tQuelleDateien.Fingerabdruck WHERE tQuelleDateien.ID IS NULL}]
		set tmp [lsort $tmp]
		foreach Element $tmp {
			lappend Protokoll(ZielDateiLoeschen) $Element
			if {$Start == 1} {
				catch {file delete $Element}
				if {[file exist $Element]} {
					tk_messageBox -title [::msgcat::mc "Fehler"] -message [::msgcat::mc "Die Datei %s kann nicht geloescht werden." $Element] -icon question -type ok -default ok
					set Abbruch 2
					break
				}
			}
		}
	}
	
	if {$Abbruch == 0} {
		# Alle Dateien aus der Quelle, die nicht im Ziel vorkommen
		# Diese Dateien müssen kopiert werden
		set Status [::msgcat::mc "in Arbeit"]
		update idletasks
		if {$Zeit > 0} {
			set Startzeit [clock seconds]
			set Stoppzeit [expr $Startzeit + 60*$Zeit]
		}
		set tmp [Datenbank eval {SELECT DISTINCT tQuelleDateien.Datei,tQuelleDateien.PfadRelativ FROM tQuelleDateien LEFT OUTER JOIN tZielDateien ON tQuelleDateien.Fingerabdruck=tZielDateien.Fingerabdruck WHERE tZielDateien.ID IS NULL}]
		set tmp2 {}
		foreach {Element Ziel} $tmp {
			lappend tmp2 "$Element -> $GlobKonf(Ziel)$Ziel"
			set Protokoll(DateiKopieren) "$Element" ; # zeigt immer die Datei an, die aktuell kopiert wird
			if {$Start == 1} {
				catch {file copy $Element $GlobKonf(Ziel)$Ziel}
				if {$Zeit > 0} {
					if {[clock seconds] > $Stoppzeit} {
						set Abbruch 1
						break
					}
				}
				if {![file exist $Element]} {
					tk_messageBox -title [::msgcat::mc "Fehler"] -message [::msgcat::mc "Die Datei %s kann nicht kopiert werden." $Element] -icon question -type ok -default ok
					set Abbruch 2
					break
				}
			}
			update
		}
		if {$Abbruch == 0} {
			set tmp2 [lsort $tmp2]
			set Protokoll(DateiKopieren) $tmp2
		}
	}
	
	# Alle Dateien, die in der Quelle und im Ziel vorhanden sind
	#puts "Dateien in Quelle und Ziel vorhanden:"
	#puts [Datenbank eval {SELECT DISTINCT tQuelleDateien.ID,tQuelleDateien.Datei,tZielDateien.ID,tZielDateien.Datei FROM tQuelleDateien JOIN tZielDateien ON tQuelleDateien.Fingerabdruck=tZielDateien.Fingerabdruck}]

	if {$Abbruch == 0} {
		# Alle Dateien, die im Ziel verschoben werden müssen
		set tmp [Datenbank eval {SELECT DISTINCT tZielDateien.Datei,tQuelleDateien.PfadRelativ FROM tQuelleDateien JOIN tZielDateien ON tQuelleDateien.Fingerabdruck=tZielDateien.Fingerabdruck WHERE tQuelleDateien.PfadRelativ <> tZielDateien.PfadRelativ}]
		set tmp2 {}
		foreach {Element Ziel} $tmp {
			lappend tmp2 "$Element -> $GlobKonf(Ziel)$Ziel"
			if {$Start == 1} {
				catch {file rename $Element $GlobKonf(Ziel)$Ziel}
				if {[file exist $Element] || (![file exist $GlobKonf(Ziel)$Ziel])} {
					tk_messageBox -title [::msgcat::mc "Fehler"] -message [::msgcat::mc "Die Datei %s kann nicht verschoben werden." $Element] -icon question -type ok -default ok
					set Abbruch 2
					break
				}
			}
		}
		if {$Abbruch == 0} {
			set tmp2 [lsort $tmp2]
			foreach Element $tmp2 {
				lappend Protokoll(DateiVerschieben) $Element
			}
		}
	}

	if {$Abbruch == 0} {
		# Alle Ordner aus dem Ziel, die nicht in der Quelle vorkommen
		# Diese Ordner müssen gelöscht werden.
		set tmp [Datenbank eval {SELECT DISTINCT tZielOrdner.OrdnerRelativ FROM tZielOrdner LEFT OUTER JOIN tQuelleOrdner ON tZielOrdner.OrdnerRelativ=tQuelleOrdner.OrdnerRelativ WHERE tQuelleOrdner.ID IS NULL}]
		set tmp [lsort $tmp]
		foreach Element $tmp {
			lappend Protokoll(ZielOrdnerLoeschen) $GlobKonf(Ziel)$Element
			if {$Start == 1} {
				catch {file delete -force $GlobKonf(Ziel)$Element}
				if {[file exist $GlobKonf(Ziel)$Element]} {
					tk_messageBox -title [::msgcat::mc "Fehler"] -message [::msgcat::mc "Der Ordner %s kann nicht geloescht werden." $GlobKonf(Ziel)$Element] -icon question -type ok -default ok
					set Abbruch 2
					break
				}
			}
		}
	}

	if {$Abbruch == 0} {
		set Status [::msgcat::mc "Fertig."]
	} elseif {$Abbruch == 1} {
		set Status [::msgcat::mc "Ende wegen Zeitueberschreitung."]
	} else {
		set Status [::msgcat::mc "Fehler."]
	}
	update idletasks
	
	. configure -cursor $Cursor
}

proc btAnalyseKlick {} {
	global GlobKonf
	global Protokoll
	
	if {$GlobKonf(Md5sum) == 1} {
		set Antwort [tk_messageBox -title [::msgcat::mc "Frage"] -message [::msgcat::mc "Sie haben Md5sum aktiviert. Dieser Vorgang dauert bei grossen bzw. vielen Dateien sehr lange. Soll Md5sum wirklich verwendet werden?"] -icon question -type yesno -default no]
		if {$Antwort eq "no"} {
			set GlobKonf(Md5sum) 0
			update idletasks
		}
	}
	
	DatenbankLeeren
	BackupStarten 0
	if {[llength $Protokoll(ZielNeueOrdner)] > 0 || [llength $Protokoll(ZielDateiLoeschen)] > 0 || [llength $Protokoll(DateiKopieren)] > 0 || [llength $Protokoll(DateiVerschieben)] > 0 || [llength $Protokoll(ZielOrdnerLoeschen)] > 0} {
		.f1.btStart configure -state normal
	}
}
	
proc btStartKlick {} {
	BackupStarten 1
	DatenbankLeeren
	.f1.btStart configure -state disabled
	update idletasks
	catch {exec sync}
	set Antwort [tk_messageBox -title [::msgcat::mc "Hinweis"] -message [::msgcat::mc "Fertig."] -icon info -type ok -default ok]
}

proc ProgrammBeenden {} {
	global GlobKonf

	catch {Datenbank close}

	if {[catch {set f [open $GlobKonf(Konfigurationsdatei) w]}] == 0} {
		puts $f "1:$GlobKonf(Quelle)"
		puts $f "2:$GlobKonf(Ziel)"
		puts $f "3:$GlobKonf(Sprache)"
		puts $f "4:$GlobKonf(Schriftgroesse)"
		puts $f "5:$GlobKonf(Md5sum)"
		close $f
	}
	exit
}

proc InfoAnzeigen {ProgrammName ProgrammTitel ProgrammVersion} {
	set Titel [::msgcat::mc "Info"]
	set Text "${ProgrammName}\nVersion ${ProgrammVersion}\n\nCopyright Oliver Scholl\nwww.tcltk.info\n
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the license, or any later version.\n
This software is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of merchantability or fitness for a particular purpose. See the GNU General Public License for more details."
	tk_messageBox -title $Titel -message $Text -type ok
}

proc SchriftgroesseAendern {Schriftgroesse} {
	global GlobSchrift
	global GlobKonf
	font configure GlobSchrift -size $Schriftgroesse
	set GlobKonf(Schriftgroesse) $Schriftgroesse
}

# ***** GUI *****

wm protocol . WM_DELETE_WINDOW {
	# wird beim Schliessen des Hauptfensters ausgeführt
	ProgrammBeenden
}

font configure GlobSchrift -size $GlobKonf(Schriftgroesse)

wm title . $GlobKonf(ProgrammTitel)


option add *Menu.tearOff 0
. configure -menu .m
menu .m
.m add command -label [::msgcat::mc "Schliessen"] -font GlobSchrift -command ProgrammBeenden
.m add cascade -label [::msgcat::mc "Sprache"] -font GlobSchrift -menu .m.sprache
.m add cascade -label [::msgcat::mc "Schriftgroesse"] -font GlobSchrift -menu .m.schriftgroesse
.m add command -label [::msgcat::mc "Info"] -font GlobSchrift -command {InfoAnzeigen $GlobKonf(ProgrammName) $GlobKonf(ProgrammTitel) $GlobKonf(ProgrammVersion)}

menu .m.sprache
.m.sprache add radiobutton -label "Deutsch" -font GlobSchrift -variable Sprache -command {SpracheAendern $Sprache}
.m.sprache add radiobutton -label "English" -font GlobSchrift -variable Sprache -command {SpracheAendern $Sprache}

menu .m.schriftgroesse
.m.schriftgroesse add radiobutton -label 8 -variable Schriftgroesse -value 8 -underline 0 -font GlobSchrift -command {SchriftgroesseAendern $Schriftgroesse}
.m.schriftgroesse add radiobutton -label 9 -variable Schriftgroesse -value 9 -underline 0 -font GlobSchrift -command {SchriftgroesseAendern $Schriftgroesse}
.m.schriftgroesse add radiobutton -label 10 -variable Schriftgroesse -value 10 -underline 0 -font GlobSchrift -command {SchriftgroesseAendern $Schriftgroesse}
.m.schriftgroesse add radiobutton -label 11 -variable Schriftgroesse -value 11 -underline 0 -font GlobSchrift -command {SchriftgroesseAendern $Schriftgroesse}
.m.schriftgroesse add radiobutton -label 12 -variable Schriftgroesse -value 12 -underline 0 -font GlobSchrift -command {SchriftgroesseAendern $Schriftgroesse}
.m.schriftgroesse add radiobutton -label 14 -variable Schriftgroesse -value 14 -underline 0 -font GlobSchrift -command {SchriftgroesseAendern $Schriftgroesse}
.m.schriftgroesse add radiobutton -label 16 -variable Schriftgroesse -value 16 -underline 0 -font GlobSchrift -command {SchriftgroesseAendern $Schriftgroesse}
.m.schriftgroesse add radiobutton -label 18 -variable Schriftgroesse -value 18 -underline 0 -font GlobSchrift -command {SchriftgroesseAendern $Schriftgroesse}
.m.schriftgroesse add radiobutton -label 20 -variable Schriftgroesse -value 20 -underline 0 -font GlobSchrift -command {SchriftgroesseAendern $Schriftgroesse}
.m.schriftgroesse add radiobutton -label 22 -variable Schriftgroesse -value 22 -underline 0 -font GlobSchrift -command {SchriftgroesseAendern $Schriftgroesse}

ttk::frame .f1
ttk::label .f1.lbQuelle -text [::msgcat::mc "Quelle:"]
ttk::label .f1.lbQuelleVerzeichnis -textvariable GlobKonf(Quelle)
ttk::button .f1.btQuelleWaehlen -text "..." -command {set GlobKonf(Quelle) [tk_chooseDirectory -initialdir $env(HOME)]}
ttk::label .f1.lbZiel -text [::msgcat::mc "Ziel:"]
ttk::label .f1.lbZielVerzeichnis -textvariable GlobKonf(Ziel)
ttk::button .f1.btZielWaehlen -text "..." -command {set GlobKonf(Ziel) [tk_chooseDirectory -initialdir $env(HOME)]} 
ttk::checkbutton .f1.cbMd5sum -text [::msgcat::mc "Md5sum verwenden"] -variable GlobKonf(Md5sum)
ttk::button .f1.btAnalyse -text [::msgcat::mc "Analyse"] -command btAnalyseKlick
ttk::button .f1.btStart -text [::msgcat::mc "Start"] -command btStartKlick -state disabled 
ttk::label .f1.lbZeit -text [::msgcat::mc "maximale Zeit (in Minuten):"]
ttk::entry .f1.enZeit -textvariable Zeit -font Schrift

ttk::notebook .n
ttk::frame .n.f1
ttk::frame .n.f2
ttk::frame .n.f3
.n add .n.f1 -text [::msgcat::mc "Hinweise"]
.n add .n.f2 -text [::msgcat::mc "Ordner"]
.n add .n.f3 -text [::msgcat::mc "Dateien"]

ttk::label .n.f1.lbProtokollQuelleDubletten -text [::msgcat::mc "Folgende Dateien kommen in der Quelle mehrfach vor:"]
ttk::frame .n.f1.f3 -relief sunken -borderwidth 1
listbox .n.f1.f3.libProtokollQuelleDubletten -height 4 -listvariable Protokoll(QuelleDubletten) -yscrollcommand ".n.f1.f3.scrolly set" -xscrollcommand ".n.f1.f3.scrollx set" -font Schrift -selectmode extended
ttk::scrollbar .n.f1.f3.scrolly -command ".n.f1.f3.libProtokollQuelleDubletten yview"
ttk::scrollbar .n.f1.f3.scrollx -orient horizontal -command ".n.f1.f3.libProtokollQuelleDubletten xview"

ttk::label .n.f1.lbProtokollZielDubletten -text [::msgcat::mc "Folgende Dateien kommen im Ziel mehrfach vor:"]
ttk::frame .n.f1.f4 -relief sunken -borderwidth 1
listbox .n.f1.f4.libProtokollZielDubletten -height 4 -listvariable Protokoll(ZielDubletten) -yscrollcommand ".n.f1.f4.scrolly set" -xscrollcommand ".n.f1.f4.scrollx set" -font Schrift -selectmode extended
ttk::scrollbar .n.f1.f4.scrolly -command ".n.f1.f4.libProtokollZielDubletten yview"
ttk::scrollbar .n.f1.f4.scrollx -orient horizontal -command ".n.f1.f4.libProtokollZielDubletten xview"

ttk::label .n.f2.lbProtokollZielNeueOrdner -text [::msgcat::mc "Folgende Ordner werden im Ziel neu angelegt:"]
ttk::frame .n.f2.f5 -relief sunken -borderwidth 1
listbox .n.f2.f5.libProtokollZielNeueOrdner -height 4 -listvariable Protokoll(ZielNeueOrdner) -yscrollcommand ".n.f2.f5.scrolly set" -xscrollcommand ".n.f2.f5.scrollx set" -font Schrift -selectmode extended
ttk::scrollbar .n.f2.f5.scrolly -command ".n.f2.f5.libProtokollZielNeueOrdner yview"
ttk::scrollbar .n.f2.f5.scrollx -orient horizontal -command ".n.f2.f5.libProtokollZielNeueOrdner xview"

ttk::label .n.f3.lbProtokollZielDateiLoeschen -text [::msgcat::mc "Folgende Dateien werden im Ziel geloescht:"]
ttk::frame .n.f3.f6 -relief sunken -borderwidth 1
listbox .n.f3.f6.libProtokollZielDateiLoeschen -height 4 -listvariable Protokoll(ZielDateiLoeschen) -yscrollcommand ".n.f3.f6.scrolly set" -xscrollcommand ".n.f3.f6.scrollx set" -font Schrift -selectmode extended
ttk::scrollbar .n.f3.f6.scrolly -command ".n.f3.f6.libProtokollZielDateiLoeschen yview"
ttk::scrollbar .n.f3.f6.scrollx -orient horizontal -command ".n.f3.f6.libProtokollZielDateiLoeschen xview"

ttk::label .n.f3.lbProtokollDateiKopieren -text [::msgcat::mc "Folgende Dateien werden kopiert:"]
ttk::frame .n.f3.f7 -relief sunken -borderwidth 1
listbox .n.f3.f7.libProtokollDateiKopieren -height 4 -listvariable Protokoll(DateiKopieren) -yscrollcommand ".n.f3.f7.scrolly set" -xscrollcommand ".n.f3.f7.scrollx set" -font Schrift -selectmode extended
ttk::scrollbar .n.f3.f7.scrolly -command ".n.f3.f7.libProtokollDateiKopieren yview"
ttk::scrollbar .n.f3.f7.scrollx -orient horizontal -command ".n.f3.f7.libProtokollDateiKopieren xview"

ttk::label .n.f3.lbProtokollDateiVerschieben -text [::msgcat::mc "Folgende Dateien werden im Ziel verschoben:"]
ttk::frame .n.f3.f8 -relief sunken -borderwidth 1
listbox .n.f3.f8.libProtokollDateiVerschieben -height 4 -listvariable Protokoll(DateiVerschieben) -yscrollcommand ".n.f3.f8.scrolly set" -xscrollcommand ".n.f3.f8.scrollx set" -font Schrift -selectmode extended
ttk::scrollbar .n.f3.f8.scrolly -command ".n.f3.f8.libProtokollDateiVerschieben yview"
ttk::scrollbar .n.f3.f8.scrollx -orient horizontal -command ".n.f3.f8.libProtokollDateiVerschieben xview"

ttk::label .n.f2.lbProtokollZielOrdnerLoeschen -text [::msgcat::mc "Folgende Ordner werden im Ziel geloescht:"]
ttk::frame .n.f2.f9 -relief sunken -borderwidth 1
listbox .n.f2.f9.libProtokollZielOrdnerLoeschen -height 4 -listvariable Protokoll(ZielOrdnerLoeschen) -yscrollcommand ".n.f2.f9.scrolly set" -xscrollcommand ".n.f2.f9.scrollx set" -font Schrift -selectmode extended
ttk::scrollbar .n.f2.f9.scrolly -command ".n.f2.f9.libProtokollZielOrdnerLoeschen yview"
ttk::scrollbar .n.f2.f9.scrollx -orient horizontal -command ".n.f2.f9.libProtokollZielOrdnerLoeschen xview"

ttk::frame .f99 -relief ridge -borderwidth 1
ttk::label .f99.lbStatus -textvariable Status

pack .f1 -side top -anchor nw -fill x
grid .f1.lbQuelle -row 0 -column 0 -sticky w -padx 3 -pady 3
grid .f1.lbQuelleVerzeichnis -row 0 -column 1 -sticky w -padx 3 -pady 3 -columnspan 2
grid .f1.btQuelleWaehlen -row 0 -column 3 -sticky w -padx 3 -pady 3
grid .f1.lbZiel -row 1 -column 0 -sticky w -padx 3 -pady 3 
grid .f1.lbZielVerzeichnis -row 1 -column 1 -sticky w -padx 3 -pady 3 -columnspan 2
grid .f1.btZielWaehlen -row 1 -column 3 -sticky w -padx 3 -pady 3

grid .f1.lbZeit -row 2 -column 0 -sticky w -padx 3 -pady 3 -columnspan 2
grid .f1.enZeit -row 2 -column 2 -sticky w -padx 3 -pady 3

grid .f1.cbMd5sum -row 2 -column 3 -sticky w -padx 3 -pady 3 -columnspan 1
grid .f1.btAnalyse -row 4 -column 0 -sticky w -padx 3 -pady 3
grid .f1.btStart -row 4 -column 1 -sticky w -padx 3 -pady 3

pack .n -side top -fill both -expand 1

pack .n.f1.lbProtokollQuelleDubletten -side top -anchor w -padx 3 -pady 3
pack .n.f1.f3 -side top -expand yes -fill both
pack .n.f1.f3.scrolly -side right -fill y
pack .n.f1.f3.scrollx -side bottom -fill x
pack .n.f1.f3.libProtokollQuelleDubletten -side left -expand yes -fill both

pack .n.f1.lbProtokollZielDubletten -side top -anchor w -padx 3 -pady 3
pack .n.f1.f4 -side top -expand yes -fill both
pack .n.f1.f4.scrolly -side right -fill y
pack .n.f1.f4.scrollx -side bottom -fill x
pack .n.f1.f4.libProtokollZielDubletten -side left -expand yes -fill both

pack .n.f2.lbProtokollZielNeueOrdner -side top -anchor w -padx 3 -pady 3
pack .n.f2.f5 -side top -expand yes -fill both
pack .n.f2.f5.scrolly -side right -fill y
pack .n.f2.f5.scrollx -side bottom -fill x
pack .n.f2.f5.libProtokollZielNeueOrdner -side left -expand yes -fill both

pack .n.f2.lbProtokollZielOrdnerLoeschen -side top -anchor w -padx 3 -pady 3
pack .n.f2.f9 -side top -expand yes -fill both
pack .n.f2.f9.scrolly -side right -fill y
pack .n.f2.f9.scrollx -side bottom -fill x
pack .n.f2.f9.libProtokollZielOrdnerLoeschen -side left -expand yes -fill both

pack .n.f3.lbProtokollDateiVerschieben -side top -anchor w -padx 3 -pady 3
pack .n.f3.f8 -side top -expand yes -fill both
pack .n.f3.f8.scrolly -side right -fill y
pack .n.f3.f8.scrollx -side bottom -fill x
pack .n.f3.f8.libProtokollDateiVerschieben -side left -expand yes -fill both

pack .n.f3.lbProtokollDateiKopieren -side top -anchor w -padx 3 -pady 3
pack .n.f3.f7 -side top -expand yes -fill both
pack .n.f3.f7.scrolly -side right -fill y
pack .n.f3.f7.scrollx -side bottom -fill x
pack .n.f3.f7.libProtokollDateiKopieren -side left -expand yes -fill both

pack .n.f3.lbProtokollZielDateiLoeschen -side top -anchor w -padx 3 -pady 3
pack .n.f3.f6 -side top -expand yes -fill both
pack .n.f3.f6.scrolly -side right -fill y
pack .n.f3.f6.scrollx -side bottom -fill x
pack .n.f3.f6.libProtokollZielDateiLoeschen -side left -expand yes -fill both

pack .f99 -side bottom -fill x
pack .f99.lbStatus -side left

# ***** Nachdem die GUI erzeugt wurde *****

DatenbankAnlegen
SpracheAendern $GlobKonf(Sprache)
